/**
 * @file
 * @brief QCD beta function
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.3
 *
 * @section LICENSE
 *
 * Copyright 2014 Andreas Maier
 *
 * This file is part of the QCD library.
 *
 * The QCD library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QCD library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QCD library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <array>
#include "colour.hpp"

#if __cplusplus >= 201402L
#define QCD_CONSTEXPR_AFTER_CXX14 constexpr
#else
#define QCD_CONSTEXPR_AFTER_CXX14
#endif

//! -
namespace QCD{

  //! Highest implemented order of the beta function
  constexpr unsigned beta_max_order = 3;

  namespace detail{
    namespace beta{
      constexpr double pi = 3.141592653589793238462643383279;
      constexpr double zeta_3 = 1.202056903159594285399738161511449990765;
      constexpr auto max_order = beta_max_order;

      // beta = - sum beta_ij nf^j as^(i+1)
      // as = alpha_s/pi
      //taken from Czakon hep-ph/0411261
      constexpr double beta_00 = 11./12.*c_A;
      constexpr double beta_01 = - 1./3.*T_f;

      constexpr double beta_10 = 17./24.*c_A*c_A;
      constexpr double beta_11 = -(c_F/4.+5./12.*c_A)*T_f;

      constexpr double beta_20 = 2857./3456.*c_A*c_A*c_A;
      constexpr double
      beta_21 = -T_f*(1415./1728.*c_A*c_A + 205./576.*c_A*c_F - c_F*c_F/32.);
      constexpr double
      beta_22 = T_f*T_f*(79./864.*c_A + 11./144.*c_F);

      constexpr double
      beta_30 =
	+ c_A*c_A*c_A*c_A*(150653./124416. - 11./576.*zeta_3)
	+ d_abcd_AA/N_A*(- 5./144. + 11./12.*zeta_3)
	;

      constexpr double
      beta_31 =
	T_f*(
	    + c_A*c_A*c_A*(-39143./20736. + 17./96.*zeta_3)
	    + c_A*c_A*c_F*(7073./62208. - 41./144.*zeta_3)
	    + c_A*c_F*c_F*(-1051./1728. + 11./72.*zeta_3)
	    + c_F*c_F*c_F*23./128.
	)
	+ d_abcd_AF/N_A*(2./9. - 13./6.*zeta_3)
	;

      constexpr double
      beta_32 =
	T_f*T_f*(
	    + c_A*c_A*(3965./10368. + 7./72.*zeta_3)
	    + c_A*c_F*(67./243. + 7./36.*zeta_3)
	    + c_F*c_F*(169./864. - 11./36.*zeta_3)
	)
	+ d_abcd_FF/N_A*(- 11./36. + 2./3.*zeta_3)
	;

      constexpr double
      beta_33 = T_f*T_f*T_f*(c_A*53./7776. + c_F*77./3888.);

      constexpr auto max_nf_pow =
	std::array<unsigned,1u + max_order>{{1u, 1u, 2u, 3u}};

      inline QCD_CONSTEXPR_AFTER_CXX14
      unsigned coeffs_start_index(unsigned order){
	return (order == 0)?
	  0:
	  (1u + max_nf_pow[order-1] + coeffs_start_index(order-1))
	  ;
      }

      constexpr auto coeffs = std::array<double,11>{{
	  beta_00, beta_01,
	  beta_10, beta_11,
	  beta_20, beta_21, beta_22,
	  beta_30, beta_31, beta_32, beta_33
	}};

      inline QCD_CONSTEXPR_AFTER_CXX14
      unsigned beta_max_nf_pow(unsigned order){
	// assert(order <= max_order);
	return max_nf_pow[order];
      }

      inline QCD_CONSTEXPR_AFTER_CXX14
      double beta_nf_coeff(unsigned order, unsigned nfpow){
	// assert(order <= max_order);
	// assert(nfpow <= max_nf_pow(order));
	return( coeffs[coeffs_start_index(order) + nfpow] );
      }

      inline QCD_CONSTEXPR_AFTER_CXX14
      double helper(unsigned order, int nf, unsigned nf_pow){
	//assert(nf_pow <= max_nf_pow[order])
	return
	  (nf_pow == max_nf_pow[order])?
	  beta_nf_coeff(order, nf_pow):
	  beta_nf_coeff(order, nf_pow) + nf*helper(order, nf, nf_pow + 1)
	  ;
      }

      inline QCD_CONSTEXPR_AFTER_CXX14
      double beta(unsigned order, int nf){
	// assert(order <= max_order);
	using namespace detail::beta;
	return helper(order, nf, 0);
      }

      inline QCD_CONSTEXPR_AFTER_CXX14
      double helper(
	  unsigned order,
	  int nf,
	  double alpha_s,
	  unsigned cur_order
      ){
	//assert(order <= max_order)
	return (cur_order == order)?
	  beta(order, nf):(
	      + beta(cur_order, nf)
	      + alpha_s/pi * helper(order, nf, alpha_s, cur_order + 1)
	  );
      }
    }
  }

  //! The QCD beta function
  /**
   * @param order   Perturbative order of the beta function
   * @param nf      Number of active flavours
   * @param alpha_s Value of the strong coupling constant
   * @returns       The value of the beta function
   *
   * The beta function is defined via the renormalisation group equation
   * \f$\beta = \frac{d\ln\alpha_s(\mu)}{d\ln \mu^2} \f$, where
   * \f$\alpha_s(\mu)\f$ is the strong coupling constant at the scale \f$\mu\f$.
   *
   */
  inline QCD_CONSTEXPR_AFTER_CXX14
  double beta(unsigned order, int nf, double alpha_s){
    // assert(order <= max_order);
    using namespace detail::beta;
    return - alpha_s/pi*helper(order, nf, alpha_s, 0);
  }

  //! Perturbative coefficients of the QCD beta function
  /**
   * @param order   Perturbative order of the coefficient
   * @param nf      Number of active flavours
   * @returns       The value of the coefficient of the beta function
   *
   * The coefficients of the beta function are given by
   * \f$\beta = - \sum_{i=0}^\infty \beta_i (\frac{\alpha_s}{\pi})^{i+1}\f$.
   * For example, beta(0, 5) returns the value of \f$\beta_0\f$
   * for 5 active flavours.
   *
   */
  inline QCD_CONSTEXPR_AFTER_CXX14
  double beta(unsigned order, int nf){
    return detail::beta::beta(order, nf);
  }

  //! Highest power of number of flavours in appearing in the beta function
  /**
   * @param order   Perturbative order of the beta function
   * @returns       The highest power of the number of active flavours
   *                in the formula for the beta function up to the given order
   */
  inline QCD_CONSTEXPR_AFTER_CXX14
  unsigned beta_max_nf_pow(unsigned order){
    return detail::beta::beta_max_nf_pow(order);
  }
  //! Coefficient of some power of number of flavours in the beta function
  /**
   * @param order   Perturbative order
   * @param nfpow   Power of number of active flavour
   * @returns       The coefficient of \f$n_f^{\rm nfpow}\f$ in the formula for \f$\beta_{\rm order}\f$
   */
  inline QCD_CONSTEXPR_AFTER_CXX14
  double beta_nf_coeff(unsigned order, unsigned nfpow){
    return detail::beta::beta_nf_coeff(order, nfpow);
  }

  // the following functions are just defined for safety, e.g.
  // to prevent unintended implicit conversion from negative ints
  // please avoid using them

  inline QCD_CONSTEXPR_AFTER_CXX14
  double beta_nf_coeff(int order, unsigned nfpow){
    // assert(order >= 0);
    return beta_nf_coeff(static_cast<unsigned>(order), nfpow);
  }

  inline QCD_CONSTEXPR_AFTER_CXX14
  double beta_nf_coeff(unsigned order, int nfpow){
    // assert(nfpow >= 0);
    return beta_nf_coeff(order, static_cast<unsigned>(nfpow));
  }

  inline QCD_CONSTEXPR_AFTER_CXX14
  double beta_nf_coeff(int order, int nfpow){
    // assert(order >= 0);
    // assert(nfpow >= 0);
    return
      beta_nf_coeff(static_cast<unsigned>(order), static_cast<unsigned>(nfpow));
  }

}

#undef QCD_CONSTEXPR_AFTER_CXX14
