#include <new>

#include "gsl/gsl_integration.h"

namespace detail{
  /**
   * Provides a function with the signature required by gsl
   * that can actually call an arbitrary Functor
   */
  template<class Functor>
    inline
    double call_from_gsl(double t, void * p){
    Functor * f = static_cast<Functor *>(p);
    return (*f)(t);
  }

  //! Class for integrating a function of type Integrand.
  /**
   * The function has to be real-valued.
   */
  template<class Integrand>
    class integrator{
    static constexpr size_t workspace_size = 200;
    Integrand i;
    gsl_integration_cquad_workspace * space;
    double abserr;
    size_t nevals;
  public:
    //! Constructor
    /**
     * @param in    The integrand function.
     */
    integrator(Integrand in):
      i(in),
      space(gsl_integration_cquad_workspace_alloc(workspace_size)){
         if(space == nullptr) throw ::std::bad_alloc();
    }
    ~integrator(){
      gsl_integration_cquad_workspace_free(space);
    }
    //! Integrate over the integrand specified during construction.
    /**
     * @param a       Lower bound of the integration.
     * @param b       Upper bound of the integration.
     * @param epsabs  Absolute accuracy
     * @param epsrel  Relative accuracy
     * @returns       The integral from a to b with the specified accuracy.
     */
    double integrate(
	double a, double b,
	double epsabs = 1e-8, double epsrel = 1e-6
    ){
      double result;
      gsl_function F;
      F.function = &detail::call_from_gsl<Integrand>;
      F.params = &i;
      gsl_integration_cquad(
	  &F, a, b, epsabs, epsrel, space, &result, &abserr, &nevals
      );
      return result;
    }
    //! Obtain absolute error of last integration
    double error() const {
      return abserr;
    }
    //! Obtain number of evaluations during last integration
    double num_evals() const {
      return nevals;
    }
  };
}

template<class F>
double integral(double lower, double higher, F f){
  auto i = detail::integrator<F>{f};
  return i.integrate(lower, higher);
}
