/**
 * @file
 * @brief Default strong coupling constants
 *
 * This header declares default coupling constants for bottom
 * (nl = 4 active flavours) and top functions.
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.3
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include "constants.hpp"

//! -
namespace QQbar_threshold{

  struct alpha_s_options;

  //! The strong coupling used for top-related functions
  /**
   * @param mu           renormalisation scale
   * @param alpha_s_mZ   strong coupling at the scale \ref mZ
   * @returns            The value of the strong coupling theory at the given scale
   *
   * Both input and output value for the strong coupling are in the
   * theory with five active flavours.
   */
  double alpha_s_top(double mu, double alpha_s_mZ = QQbar_threshold::alpha_s_mZ);

  //! The strong coupling used for top-related functions
  /**
   * @param mu           renormalisation scale
   * @param alpha_s_mu0  a pair {mu0, alpha_s} of a reference scale mu0 and the strong coupling at the scale mu0
   * @returns            The value of the strong coupling at the given renormalisation scale
   *
   * Both input and output value for the strong coupling are in the
   * theory with five active flavours.
   *
   */
  double alpha_s_top(double mu, alpha_s_options alpha_s_mu0);

  //! The strong coupling used for bottom-related functions
  /**
   * @param mu           renormalisation scale
   * @param alpha_s_mZ   strong coupling at the scale \ref mZ
   * @returns            The value of the strong coupling theory at the given scale
   *
   * For the input value five active flavours are assumed.
   * The output value is given in the four-flavour theory.
   * Decoupling is performed at the scale \ref mu_thr assuming
   * a scale-invariant bottom mass \ref mb_SI.
   */
  double alpha_s_bottom(double mu, double alpha_s_mZ = QQbar_threshold::alpha_s_mZ);

  //! The strong coupling used for bottom-related functions
  /**
   * @param mu           renormalisation scale
   * @param alpha_s_mu0  a pair {mu0, alpha_s} of a reference scale mu0 and the strong coupling at the scale mu0
   * @returns            The value of the strong coupling at the given renormalisation scale
   *
   * For the input value, five active flavours are assumed
   * if mu0 is above the bottom decoupling threshold \ref mu_thr,
   * and four active flavours if mu0 is below \ref mu_thr.
   * The output value is always given in the four-flavour theory.
   * If necessary, decoupling is performed at the scale \ref mu_thr
   * assuming a scale-invariant bottom mass \ref mb_SI.
   */
  double alpha_s_bottom(double mu, alpha_s_options alpha_s_mu0);
}
