/**
 * @file
 * @brief Functions to calculate \f$q\bar{q}\f$ bound state energy levels
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.3
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */

#pragma once

#include "parameters.hpp"

//! -
namespace QQbar_threshold{

  //! Computes the binding energy \f$E_n\f$ of an S-wave \f$t\bar{t}\f$ bound state
  /**
   *
   * @param n     principal quantum number
   * @param mu    renormalisation scale
   * @param m     top quark mass
   * @param o     perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt   further options
   * @return      The binding energy \f$E_n\f$.
   *
   * The bound state mass
   * is given by \f$2m_t+E_n\f$, where \f$m_t\f$
   * is the mass of the top quark in the given scheme.
   *
   * The supported options and their default values are:
   *   - options::mass_scheme = {\ref PSshift, mu_f_top}
   *   - options::alpha_s_mZ = #alpha_s_mZ
   *   - options::contributions = (all set to 1)
   *   - options::beyond_QCD = SM::all
   *   - options::alpha = #alpha_mZ
   *   - options::m_Higgs = #m_Higgs
   *   - options::Yukawa_factor = 1
   *   - options::r4 = 824.12 (for nl == 5)
   *   - options::ml = 0
   *   - options::double_light_insertion = false
   *
   * @see options, bbbar_energy_level, ttbar_residue
   */
  double ttbar_energy_level(
      int n, double mu, double m, order o,
      options opt = top_options()
  );

  //! Computes the binding energy \f$E_n\f$ of an S-wave \f$b\bar{b}\f$ bound state
  /**
   *
   * @param n     principal quantum number
   * @param mu    renormalisation scale
   * @param m     bottom quark mass
   * @param o     perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt   further options
   * @return      The binding energy \f$E_n\f$.
   *
   * The bound state mass
   * is given by \f$M_{\Upsilon(nS)} = 2m_b+E_n\f$, where \f$m_b\f$
   * is the mass of the bottom quark in the given scheme.
   *
   * The supported options and their default values are:
   *   - options::mass_scheme = {\ref PSshift, mu_f_bottom}
   *   - options::alpha_s_mZ = #alpha_s_mZ
   *   - options::contributions = (all set to 1)
   *   - options::beyond_QCD = SM::QED
   *   - options::alpha = #alpha_Y
   *   - options::m_Higgs = #m_Higgs
   *   - options::Yukawa_factor = 1
   *   - options::r4 = 1220.3 (for nl == 4)
   *   - options::ml = 0
   *   - options::double_light_insertion = false
   *
   * @see options, ttbar_energy_level, bbbar_residue
   */
  double bbbar_energy_level(
      int n, double mu, double m, order o,
      options opt = bottom_options()
  );

  namespace detail{
    /** \internal
     * Main function for binding energy
     * @param n    Principal quantum number.
     * @param p    Further parameters.
     * @param m    Unshifted (input) mass.
     * @return     The binding energy
     */
    double QQbar_energy_level(int n, parameters const & p, double m_unshifted);

    /** \internal
     * Helper function for binding energy
     * Implements the part without mass shift
     * @param n    Principal quantum number.
     * @param p    Further parameters.
     * @return     The binding energy
     */
    double QQbar_energy_level_no_shift(int n, parameters const & p);

    /** \internal
     * Leading-order binding energy
     * @param n    Principal quantum number.
     * @param p    Further parameters.
     * @return     The binding energy
     */
    double E_LO(int n, parameters const & p);

    // higher order corrections
    // E = E_LO * (1 + e1 + ... + e<order>)
    double e1(int n, parameters const & p);
    double e2(int n, parameters const & p);
    double e3(int n, parameters const & p);
    double e1_no_mass_shift(int n, parameters const & p);
    double e2_no_mass_shift(int n, parameters const & p);
    double e3_no_mass_shift(int n, parameters const & p);

    //charm integrand
    complex light_integrand_NNLO(complex const & u, parameters const & p);

  }
}
