/**
 * @file
 * @brief Resum poles into the Green functions
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.3
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <complex>

#include "parameters.hpp"

namespace QQbar_threshold{

  namespace detail{
    typedef std::complex<double> complex;
    //! Calculate the pole-resummation correction to the polarisation function
    /**
     * @param n    Principal quantum number
     * @param p    Further parameters
     * @returns    The pole-resummation correction
     *
     * The correction is given by
     * \f$\frac{4m_q^2}{s}\bigg(\frac{\big[Z_n\big]_{\rm expanded}}{\big[E_n - E\big]_{\rm unexpanded}} - \bigg[\frac{Z_n}{E_n - E}\bigg]_{\rm expanded}\bigg)$\f
     * This function is used for the vector polarisation function.
     */
    complex resum_pole_Pi(int n, parameters const & p);

    //! Calculate the pole-resummation correction to the Green function
    /**
     * @tparam O   Perturbative order of the Green function correction
     * @param n    Principal quantum number
     * @param p    Further parameters
     * @returns    The pole-resummation correction
     *
     * The correction is given by the Oth order coefficient of
     * \f$\frac{\big[|\psi_n(0)|^2\big]_{\rm expanded}}{\big[E_n - E\big]_{\rm unexpanded}} - \bigg[\frac{|\psi_n(0)|^2}{E_n - E}\bigg]_{\rm expanded}$\f
     * This function is used for pole resummation in the electroweak part.
     */
    template<int O>
    complex resum_pole_GF(int n, parameters const & p);
  }
}
