/**
 * @file
 * @brief Conversion between mass schemes
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.3
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <exception>
#include <string>

#include "mass_schemes.hpp"
#include "parameters.hpp"

#include "QCD/colour.hpp"
#include "QCD/beta.hpp"

namespace QQbar_threshold{

  //! Computes the pole mass from a bottom quark mass in a different scheme
  /**
   * @param mu  renormalisation scale
   * @param m   bottom quark mass
   * @param o   perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt further options
   * @returns   the bottom pole mass
   *
   * The supported options and their default values are:
   *   - options::mass_scheme = {\ref PSshift, mu_f_bottom}
   *   - options::alpha_s_mZ = #alpha_s_mZ
   *   - options::contributions = (all set to 1)
   *   - options::beyond_QCD = SM::QED
   *   - options::alpha = #alpha_Y
   *   - options::m_Higgs = #m_Higgs
   *   - options::Yukawa_factor = 1
   *   - options::r4 = 1220.3 (for nl == 4)
   *   - options::ml = 0
   *   - options::double_light_insertion = false
   *
   * #### Example
   *
   * This computes the pole mass for a PS bottom quark mass of 4.5 GeV
   * at a scale of 5 GeV at NNNLO:
   * @code
   * options opt = bottom_options();
   * opt.mass_scheme = {PSshift, 2};
   * double mb_PS = 4.5;
   * double mb_pole = pole_bottom_mass(5, mb_PS, N3LO, opt);
   * @endcode
   *
   * @see mass_schemes.hpp
   */
  double bottom_pole_mass(
      double mu, double m, order o,
      options opt = bottom_options()
  );

  //! Computes the pole mass from a top quark masses in a different scheme
  /**
   * @param mu  renormalisation scale
   * @param m   top quark mass
   * @param o   perturbative order (LO, NLO, N2LO, or N3LO)
   * @param opt further options
   * @returns   the top pole mass
   *
   * The supported options and their default values are:
   *   - options::mass_scheme = {\ref PSshift, mu_f_top}
   *   - options::alpha_s_mZ = #alpha_s_mZ
   *   - options::contributions = (all set to 1)
   *   - options::beyond_QCD = SM::all
   *   - options::alpha = #alpha_mZ
   *   - options::m_Higgs = #m_Higgs
   *   - options::Yukawa_factor = 1
   *   - options::r4 = 824.12 (for nl == 5)
   *   - options::ml = 0
   *   - options::double_light_insertion = false
   *
   * #### Example
   *
   * This computes the pole mass for a \f$\overline{\rm MS}\f$ top quark mass
   * \f$m_t(m_t) = 162\f$ GeV at N3LO:
   * @code
   * options opt = top_options();
   * double mt_mt = 162;
   * opt.mass_scheme = {MSbar, mt_mt};
   * double mt_pole = pole_top_mass(mt_mt, mt_mt, N3LO, opt);
   * @endcode
   *
   * @see mass_schemes.hpp
   */
  double top_pole_mass(
      double mu, double m, order o,
      options opt = top_options()
  );


  namespace detail{
    //! \internal
    /** Shifts with respect to the Pole mass for different mass schemes
     *  m_Pole = m_scheme + sum_{i=0}^\\infty delta_m<scheme, i>
     */
    double delta_m(int order, parameters const & p);

    template<scheme_name S>
    double delta_m(int order, parameters const & p);

    template<scheme_name S, int order>
    double delta_m(parameters const & p);

    template<scheme_name S, int order>
    double delta_m(parameters const & p);

    double mass_shift(parameters const & p);
    double bottom_mass_shift(
        double mu, double m, order o,
        options opt = bottom_options()
    );
    double top_mass_shift(
        double mu, double m, order o,
        options opt = top_options()
    );
  }

}
