/**
 * @file
 * @brief Contains definition of the threshold class
 *
 * \internal
 * The idea is to construct an "unevaluated" object, e.g. threshold() + 5.0
 * which is related to the center-of-mass energy of some production threshold.
 * This is used as an argument to a xsection function. The function then has
 * all necessary information (quark flavour, scale, ...) to evaluate the
 * this expression numerically calling the value method and providing the
 * actual threshold center-of-mass energy as the argument.
 *
 * This is implemented using CRTP and expression templates.
 * \endinternal
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.3
 *
 * @section LICENSE
 *
 * Copyright 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once
#include <memory>

#include "parameters.hpp"

//-
namespace QQbar_threshold{
  namespace detail{
    /** \internal
     * Base class for expression templates.
     * @tparam E  When deriving from this class, this has to be set to the type
     *            of the derived class
     */
    template<typename E>
    struct expression{
      /** \internal
       * Obtain the center-of-mass energy associated with this
       * threshold expression
       */
      double value(double thr) const{
	return static_cast<E const *>(this)->value(thr);
      }

      /** \internal
       * Cast to derived
       */
      operator E&(){
	return static_cast<E&>(*this);
      }

      operator E const &() const{
	return static_cast<E const &>(*this);
      }
    };

    /** \internal
     * Class representing the difference of two expressions
     * TODO: The usage of std::shared_ptr is non-standard,
     *       but there appear to be lifetime issues with const references
     */
    template<typename E1, typename E2>
    class difference: public expression<difference<E1,E2>>{
      std::shared_ptr<E1> x;
      std::shared_ptr<E2> y;
    public:
      difference(expression<E1> const & e1, expression<E2> const & e2)
	: x(std::make_shared<E1>(e1)), y(std::make_shared<E2>(e2)) {}

      double value(double thr) const{
	return x->value(thr) - y->value(thr);
      }
    };

    /** \internal
     * Class representing the sum of two expressions
     */
    template<typename E1, typename E2>
    class sum: public expression<sum<E1,E2>>{
      std::shared_ptr<E1> x;
      std::shared_ptr<E2> y;
    public:
      sum(expression<E1> const & e1, expression<E2> const & e2)
	: x(std::make_shared<E1>(e1)), y(std::make_shared<E2>(e2)) {}

      double value(double thr) const{
	return x->value(thr) + y->value(thr);
      }
    };

    /** \internal
     * Class representing the product of two expressions
     */
    template<typename E1, typename E2>
    class product: public expression<product<E1,E2>>{
      std::shared_ptr<E1> x;
      std::shared_ptr<E2> y;
    public:
      product(expression<E1> const & e1, expression<E2> const & e2)
	: x(std::make_shared<E1>(e1)), y(std::make_shared<E2>(e2)) {}

      double value(double thr) const{
	return x->value(thr) * y->value(thr);
      }
    };

    /** \internal
     * Class representing the ratio of two expressions
     */
    template<typename E1, typename E2>
    class ratio: public expression<ratio<E1,E2>>{
      std::shared_ptr<E1> x;
      std::shared_ptr<E2> y;
    public:
      ratio(expression<E1> const & e1, expression<E2> const & e2)
	: x(std::make_shared<E1>(e1)), y(std::make_shared<E2>(e2)) {}

      double value(double thr) const{
	return x->value(thr) / y->value(thr);
      }
    };

    /** \internal
     * Class representing minus an expression (additive inverse)
     */
    template<typename E>
    class neg: public expression<neg<E>>{
      std::shared_ptr<E> x;
    public:
      explicit neg(expression<E> const & e): x(std::make_shared<E>(e)) {}

      double value(double thr) const{
	return -x->value(thr);
      }
    };

    /** \internal
     * Wrapper around a double, so we can use it more easily to construct
     * expressions containing an unevaluated threshold object.
     */
    class double_expr: public expression<double_expr>{
      double const x;
    public:
      explicit double_expr(double d): x(d) {}

      double value(double) const{
	return x;
      }
    };

    //arithmetic operations ...
    template<typename E1, typename E2>
    sum<E1, E2> operator+(expression<E1> const & e1, expression<E2> const & e2){
      return sum<E1, E2>{e1, e2};
    }

    template<typename E1, typename E2>
    difference<E1, E2> operator-(expression<E1> const & e1, expression<E2> const & e2){
      return difference<E1, E2>{e1, e2};
    }

    template<typename E1, typename E2>
    product<E1, E2> operator*(expression<E1> const & e1, expression<E2> const & e2){
      return product<E1, E2>{e1, e2};
    }

    template<typename E1, typename E2>
    ratio<E1, E2> operator/(expression<E1> const & e1, expression<E2> const & e2){
      return ratio<E1, E2>{e1, e2};
    }

    template<typename E>
    neg<E> operator-(expression<E> const & e){
      return neg<E>(e);
    }

    template<typename E1>
    difference<E1, double_expr> operator-(expression<E1> const & e1, double x){
      return e1 - double_expr{x};
    }

    template<typename E1>
    sum<E1, double_expr> operator+(expression<E1> const & e1, double x){
      return e1 + double_expr{x};
    }

    template<typename E1>
    product<E1, double_expr> operator*(expression<E1> const & e1, double x){
      return e1 * double_expr{x};
    }

    template<typename E1>
    ratio<E1, double_expr> operator/(expression<E1> const & e1, double x){
      return e1 / double_expr{x};
    }

    template<typename E2>
    sum<double_expr, E2> operator+(double x, expression<E2> const & e2){
      return double_expr{x} + e2;
    }

    template<typename E2>
    difference<double_expr, E2> operator-(double x, expression<E2> const & e2){
      return double_expr{x} - e2;
    }

    template<typename E2>
    product<double_expr, E2> operator*(double x, expression<E2> const & e2){
      return double_expr{x} * e2;
    }

    template<typename E2>
    ratio<double_expr, E2> operator/(double x, expression<E2> const & e2){
      return double_expr{x} / e2;
    }

  }

  //! Center-of-mass energy of a quark threshold
  /**
   * When used in the first argument of the cross section functions
   * objects of this class represent the naive production threshold.
   *
   * @code
   * ttbar_xsection(threshold() + 0.5, mu, mt, width, order, options)
   * @endcode
   *
   * is equivalent to
   *
   * @code
   * double sqrt_s = ttbar_threshold(mu, mt, order, options) + 0.5;
   * ttbar_xsection(sqrt_s, mu, mt, width, order, options);
   * @endcode
   *
   * @see bbbar_threshold, bbbar_xsection, ttbar_threshold, ttbar_xsection
   */
  struct threshold : public detail::expression<threshold>{
    double value(double thr) const{
      return thr;
    }
  };



}
