/**
 * @file
 * @brief Implementation of energy levels
 *
 * @see energy_levels.hpp
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.3
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include "energy_levels.hpp"

#include <functional>
#include <cmath>

#include "QCD/colour.hpp"
#include "QCD/beta.hpp"
#include "alpha_s.hpp"
#include "Wilson_coefficients.hpp"
#include "harmonic_sums.hpp"
#include "zeta.hpp"
#include "psi.hpp"
#include "Gamma.hpp"
#include "scheme_conversion.hpp"
#include "integrate.hpp"
#include "log_n.hpp"
#include "h.hpp"
#include "residues.hpp"

namespace QQbar_threshold{
  using QCD::c_F;
  using QCD::c_A;
  using QCD::T_f;
  using QCD::beta;
  using harmonic_sums::harmonic_sum;
  using namespace zeta;
  using namespace complex;
  using Gamma::Gamma;
  using psi::psi;

  double ttbar_energy_level(
      int n, double mu, double m, order o,
      options opt
  ){
    return detail::QQbar_energy_level(n, gen_top_parameters(mu, m, o, opt), m);
  }

  double bbbar_energy_level(
      int n, double mu, double m, order o,
      options opt
  ){
    return detail::QQbar_energy_level(n, gen_bottom_parameters(mu, m, o, opt), m);
  }

  namespace detail{
    // leading order binding energy
    double E_LO(int n, parameters const & p){
      return -p.m*pow(c_F*p.alpha_s/(2*n),2);
    }

    double e1(int n, parameters const & p){
      if(is_threshold_mass(p.mass_scheme)){
	return
	  + 2*delta_m(1, p)/E_LO(n, p)
	  + e1_no_mass_shift(n, p)
	  ;
      }
      return e1_no_mass_shift(n, p);
    }

    namespace{
      //! Insertion of NLO Coulomb potential
      double delta_e1_Coulomb(int n, parameters const & p);
      //! Insertion of massive light-quark correction Coulomb potential
      double delta_e1_light(int n, parameters const & p);
      double delta_e1_light(int n, double xi_n);
    }

    double e1_no_mass_shift(int n, parameters const & p){
      return p.contributions.v_Coulomb[0] * p.alpha_s/pi*(
	  + delta_e1_Coulomb(n, p)
	  + delta_e1_light(n, p)
      );
    }

    namespace{

      //! Insertion of NNLO Coulomb potential
      double delta_e2_Coulomb(int n, parameters const & p);
      //! Double insertion of NLO Coulomb potential
      double delta_e2_double_Coulomb(int n, parameters const & p);
      //! Insertion of massive light-quark correction Coulomb potential
      double delta_e2_light(int n, parameters const & p);
      //! Double insertion of massive light quark and "massless" Coulomb potential
      double delta_e2_Coulomb_light(int n, parameters const & p);
      //! Double insertion of massive light quark correction Coulomb potential
      double delta_e2_double_light(int n, parameters const & p);
    }

    double e2_no_mass_shift(int n, parameters const & p){
      using namespace Wilson_coefficients;
      return
	+ p.contributions.v_Coulomb[1]*(
	    delta_e2_Coulomb(n, p) + delta_e2_light(n, p)
	)
	+ pow(p.contributions.v_Coulomb[0], 2)*(
	    + delta_e2_double_Coulomb(n, p)
	    + (
		p.double_light_insertion?(
		    + delta_e2_Coulomb_light(n, p)
		    + delta_e2_double_light(n, p)
		):0
	    )
	)
	+ p.contributions.v_delta[0]*(
	    pow(p.alpha_s,2)*2*v0(p.nl)*pow(c_F,2)/n
	)
	+ p.contributions.v_r2inv[0]*(
	    -c_F*pow(p.alpha_s,2)*b<1>(p.nl)/n
	)
	+ p.contributions.v_p2[0]*(
	    pow(p.alpha_s*c_F,2)*(4*n - 1)/(2*pow(n,2))
	)
	+ p.contributions.v_kinetic[0]*(
	    pow(p.alpha_s*c_F,2)*(8*n - 3)/(16*pow(n,2))
	)
	;
    }

    double e2(int n, parameters const & p){
      double result = e2_no_mass_shift(n, p);
      if(is_threshold_mass(p.mass_scheme)){
	result +=
	  + 2*delta_m(2, p)/E_LO(n, p)
	  + delta_m(0, p)/p.m
	  ;
      }
      return result;
    }

    namespace{
      //! Insertion of NNNLO Coulomb potential
      double delta_e3_Coulomb(int n, parameters const & p);
      //! Double insertion of NNLO and NLO Coulomb potential
      double delta_e3_double_Coulomb(int n, parameters const & p);
      //! Triple insertion of NLO Coulomb potential
      double delta_e3_triple_Coulomb(int n, parameters const & p);
      //! Double insertion of NLO Coulomb and delta (1/m^2) potential
      double delta_e3_Coulomb_delta(int n, parameters const & p);
      //! Double insertion of NLO Coulomb and 1/r^2 (1/m) potential
      double delta_e3_Coulomb_r2inv(int n, parameters const & p);
      //! Double insertion of NLO Coulomb and momentum-dependent potential
      double delta_e3_Coulomb_p2(int n, parameters const & p);
      //! Double insertion of NLO Coulomb potential and kinetic energy correction
      double delta_e3_Coulomb_kinetic(int n, parameters const & p);
      //! Insertion of correction to delta (1/m^2) potential
      double delta_e3_delta(int n, parameters const & p);
      //! Insertion of correction to 1/r^2 (1/m) potential
      double delta_e3_r2inv(int n, parameters const & p);
      //! Insertion of correction to momentum-dependent potential
      double delta_e3_p2(int n, parameters const & p);
      //! Ultrasoft correction
      double delta_e3_ultrasoft(int n, parameters const & p);
      //! Insertion of Higgs potential
      double delta_e3_Higgs(int n, parameters const & p);
    }

    double e3_no_mass_shift(int n, parameters const & p){
      return
	+ p.contributions.v_Coulomb[2] * delta_e3_Coulomb(n, p)
	+ p.contributions.v_Coulomb[0]*p.contributions.v_Coulomb[1] * delta_e3_double_Coulomb(n, p)
	+ pow(p.contributions.v_Coulomb[0], 3) * delta_e3_triple_Coulomb(n, p)
	+ p.contributions.v_Coulomb[0]*p.contributions.v_delta[0] * delta_e3_Coulomb_delta(n, p)
	+ p.contributions.v_Coulomb[0]*p.contributions.v_r2inv[0] * delta_e3_Coulomb_r2inv(n, p)
	+ p.contributions.v_Coulomb[0]*p.contributions.v_p2[0] * delta_e3_Coulomb_p2(n, p)
	+ p.contributions.v_Coulomb[0]*p.contributions.v_kinetic[0] * delta_e3_Coulomb_kinetic(n, p)
	+ p.contributions.v_delta[1] * delta_e3_delta(n, p)
	+ p.contributions.v_r2inv[1] * delta_e3_r2inv(n, p)
	+ p.contributions.v_p2[1] * delta_e3_p2(n, p)
	+ p.contributions.ultrasoft[0] * delta_e3_ultrasoft(n, p)
	+ p.contributions.v_Higgs[0] * delta_e3_Higgs(n, p)
	;
    }

    double e3(int n, parameters const & p){
      double result = e3_no_mass_shift(n, p);
      if(is_threshold_mass(p.mass_scheme)){
	result +=
	  + 2*delta_m(3, p)/E_LO(n, p)
	  + delta_m(1, p)/p.m
	  + delta_m(0, p)*(
	      + e1_no_mass_shift(n, p)
	      - p.contributions.v_Coulomb[0]*p.alpha_s/pi*4*beta(0, p.nl)
	  )/p.m;
      }
      return result;
    }

    double QQbar_energy_level_no_shift(int n, parameters const & p){
      // E = E_LO * (1 + e1 + ... + e<order>)
      double e_factor = 1;
      switch(p.order){
      case 3: e_factor += e3(n, p);
      case 2: e_factor += e2(n, p);
      case 1: e_factor += e1(n, p);
      default:;
      }
      return e_factor * E_LO(n, p);
    }

    double QQbar_energy_level(int n, parameters const & p, double m_unshifted){
      double E_N = QQbar_energy_level_no_shift(n, p)
        /*
         * add mass shift correction, so that the bound state position
         * 2*m_unshifted + QQbar_energy_level
         * is scheme-independent
         */
        + 2*(p.m - m_unshifted)
        ;
      if(is_threshold_mass(p.mass_scheme)){
        E_N += 2*delta_m(0, p);
      }
      return E_N;
    }

    namespace{

      /***************************************************
       *        NLO corrections to binding energy        *
       ***************************************************/
      double delta_e1_Coulomb(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	return
	  + a1_eff(p)/2
	  + 4*beta(0, p.nl)*( log_n(n, p) + harmonic_sum<1>(n) )
	;
      }

      /************************* NLO light-quark correction *********************
       *                                                                        *
       * NLO massive light-quark correction to the binding energy.              *
       * Equation numbers refer to                                              *
       * Beneke, Maier, Piclum, Rauh, Nucl. Phys. B891 42-72, arXiv:1411.3132	*
       * \cite Beneke:2014pta							*
       *									*
       **************************************************************************/

      //! eta helper function
      /**
       * See eq. (C.12)
       * The sum over i (cf. eqs. (C.7), (C.8)) is already included here
       */
      complex eta_light(complex const & u, int n){
	assert(n > 0);
	/* possible optimisations:                                      */
	/* - use recursion of gammas                                    */
	/* - precompute u-independent part (typically, n will be small) */
	complex result = 0;
	for(int i = 1; i <= n; ++i){
	  result += Gamma(u+i)/
	    (Gamma(n-i+1)*pow(Gamma(i),2)*Gamma(i+1)*Gamma(u+2-i));
	}
	result *= Gamma(-u)*Gamma(u+2);
	return result;
      }

      //! Integrand of light-quark correction to e1
      /**
       * See eq. (C.7)
       */
      complex e1_light_integrand(int n, complex const & u, double xi){
	return eta_light(u, n)*Gamma(2-u/2)/(u*Gamma((5-u)/2)*pow(xi,u));
      }

      /* real part of the integrand after transforming the */
      /* integration variable to real values               */
      double re_e1_integrand(int n, double t, double xi){
	static constexpr double del = 0.5;
	auto u = complex{del, tan(pi/2*t)};
	return real(e1_light_integrand(n, u, xi))/pow(cos(pi/2*t),2);
      }

      //! Massive light-quark correction to e1
      /**
       * See eq. (C.7)
       *
       * e^{(1)}_{m_c} corresponds to alpha_s/pi*delta_e1_light
       */
      double delta_e1_light(int n, parameters const & p){
	if(p.ml == 0) return 0;
	double const xi_n = sqrt(- E_LO(n,p)*p.m)/p.ml;
	return detail::delta_e1_light(n, xi_n);
      }

      double delta_e1_light(int n, double xi_n){
	using namespace std::placeholders;
	auto integrand = std::bind(re_e1_integrand, n, _1, xi_n);
	auto i = integrate::make_integrator(integrand);
	return -Gamma(n)*sqrt(pi)/8*i.integrate(-1, 1);
      }

      /***************************************************
       *       NNLO corrections to binding energy        *
       ***************************************************/

      /* correction to e2 from one insertion of coulomb potential */
      double delta_e2_Coulomb(int n, parameters const & p){
	using Wilson_coefficients::a;
	return pow(p.alpha_s/pi,2)/(8*n)*(
	    2*(
		2*a<1>(p.nl)*n*4*beta(0, p.nl)
		+ 4*(-1 + log_n(n, p)*n)*pow(4*beta(0, p.nl),2)
		+ n*16*beta(1, p.nl)
	    )*harmonic_sum<1>(n)
	    + 4*n*pow(4*beta(0, p.nl)*harmonic_sum<1>(n),2)
	    + n*(
		a<2>(p.nl)
		+ 4*a<1>(p.nl)*log_n(n, p)*4*beta(0, p.nl)
		+ 4*pow(4*beta(0, p.nl)*log_n(n, p),2)
		+ pow(4*beta(0, p.nl)*pi,2)
		+ 2*log_n(n, p)*16*beta(1, p.nl)
		+ 4*pow(4*beta(0, p.nl),2)*harmonic_sum<2>(n)
	    )
	);
      }

      /* correction to e2 from two insertions of coulomb potential */
      double delta_e2_double_Coulomb(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	return pow(p.alpha_s/pi,2)/(48*n)*(
	    12*4*beta(0, p.nl)*(
		+ a1_eff(p)*n
		+ 2*(1 +(-1 + log_n(n, p))*n)*4*beta(0, p.nl)
	    )*harmonic_sum<1>(n)
	    + 12*n*pow(4*beta(0, p.nl)*harmonic_sum<1>(n),2)
	    + n*(
		+ 3*pow(a1_eff(p),2)
		- 12*a1_eff(p)*4*beta(0, p.nl) * (1 - log_n(n, p))
		- 24*log_n(n, p)*pow(4*beta(0, p.nl),2)
		+ 12*pow(4*beta(0, p.nl)*log_n(n, p),2)
		+ 24*pow(4*beta(0, p.nl),2)*(harmonic_sum<2>(n)-zeta_2)
		- 24*n*pow(4*beta(0, p.nl),2)*(harmonic_sum<3>(n)-zeta_3)
	    )
	);
      }

      /************************* NNLO light-quark correction *********************
       *                                                                        *
       * NNLO massive light-quark correction to the binding energy.              *
       * Equation numbers refer to                                              *
       * Beneke, Maier, Piclum, Rauh, Nucl. Phys. B891 42-72, arXiv:1411.3132	*
       * \cite Beneke:2014pta							*
       *									*
       **************************************************************************/

      /* naive implementation of cotangent */
      complex cot(complex z){
	using std::tan;
	return 1/tan(z);
      }
    }

    //! Common integrand for NNLO light-quark bound-state corrections
    /**
     * See eq. (C.15)
     * I^{(2)}(u) corresponds to 1/12*light_integrand_NNLO
     */
    complex light_integrand_NNLO(complex const & u, parameters const & p){
      /* Parameters for the charm effects in the NNLO Coulomb potential */
      /* Source: A. Hoang, hep-ph/0008102 */
      /* TODO: code duplication (cmp. mass_shifts.cpp) */
      static constexpr double c2 = 0.470;
      static constexpr double d2 = 1.120;
      static constexpr double c1 = -0.8316040946513316;
      static constexpr double d1 = 1.8316040946513317;
      using Wilson_coefficients::a1_eff;

      return
	+ 1./6.*Gamma(3-u/2)/((u-2)*Gamma((7-u)/2))
	- 1./12.*(
	    3*a1_eff(p) - 6*log(p.ml/p.mu)*4*beta(0, p.nl)
	    + (
		2/u + psi<0>(2-u/2) - psi<0>((5-u)/2)
		- log(4) + pi*cot(pi/2*u)
	    )*(2 + 3*4*beta(0, p.nl))
	)*Gamma(2-u/2)/(u*Gamma((5-u)/2))
	+ (
	    - 1./18.*Gamma((u-5)/2)/((u-6)*Gamma(u/2-2))
	    - 1./6.*Gamma((u-3)/2)/((u-4)*Gamma(u/2-1))
	    + 2./9.*Gamma((u+1)/2)/(u*Gamma(u/2+1))
	)*cot(pi/2*u)
	- 19./6./sqrt(pi)*(c1*pow(c2,u) + d1*pow(d2,u))/u
	;
    }

    namespace{
      //! Integrand for NNLO light-quark correction to binding energy
      /**
       * The second comes part from using the MSbar scheme
       * for the light-quark mass.
       * See eq (C.16)
       */
      complex e2_light_integrand(int n, complex const & u, double xi, parameters const & p){
	return eta_light(u, n)*pow(xi,-u)*(
	    detail::light_integrand_NNLO(u, p) -
	    (2./3. + log(p.mu/p.ml))*Gamma(2 - u/2)/Gamma((5 - u)/2)
	);
      }

      /* real part of the integrand after transforming the */
      /* integration variable to real values               */
      double re_e2_integrand(int n, double t, double xi, parameters const & p){
	static constexpr double del = 0.5;
	auto u = complex{del, tan(pi/2*t)};
	return real(e2_light_integrand(n, u, xi, p))/pow(cos(pi/2*t),2);
      }

      //! Single massive light-quark correction to e2
      /**
       * See eqs. (C.8), (C.16)
       *
       * e^{(2)}_{m_c} corresponds to delta_e2_light (no extra factors here)
       */
      double delta_e2_light(int n, parameters const & p){
	using namespace std::placeholders;
	if(p.ml == 0) return 0;
	double const xi_n = sqrt(- E_LO(n,p)*p.m)/p.ml;
	auto integrand = std::bind(re_e2_integrand, n, _1, xi_n, p);
	auto i = integrate::make_integrator(integrand);
	return pow(p.alpha_s/pi,2)*Gamma(n)*sqrt(pi)/4*i.integrate(-1, 1);
      }

      /* double light mass correction to e2 */
      static constexpr int s_max = 10;

      //! Double massive light-quark correction to e2
      /**
       * See eq. (C.18)
       *
       * e^{(2)}_{m_c, m_c} corresponds exactly to delta_e2_double_light
       *
       * @see h.hpp
       */
      double delta_e2_double_light(int n, parameters const & p){
	if(p.ml == 0) return 0;
	double const xi_n = sqrt(- E_LO(n,p)*p.m)/p.ml;
	double result = 0;
	for(int s = 1; s < s_max; ++s){
	  if(s == n) continue;
	  result += pow(h_light<0>(n,  s, xi_n), 2)/(s*(s - n));
	}
	result *= 2;
	result += h_light<0>(n,  n, xi_n)/(n*n)*(
	    h_light<0>(n,  n, xi_n)*(7+4*n*psi<0>(n)) -
	    4*h_light<1>(n,  n, xi_n)
	);
	result *= pi/pow(4*Gamma(n),2);
	return  pow(p.alpha_s/pi,2)*(
	    result - delta_e1_light(n, xi_n)*delta_f1_light(n, xi_n)
	);
      }

      //! Mixed light mass/Coulomb correction to e2
      /**
       * See eq. (C.24)
       *
       * e^{(2)}_{m_c, massless} corresponds exactly to delta_e2_Coulomb_light
       *
       * @see h.hpp
       */
      double delta_e2_Coulomb_light(int n, parameters const & p){
	if(p.ml == 0) return 0;
	double const xi_n = sqrt(- E_LO(n,p)*p.m)/p.ml;
	double result = 0;
	for(int s = 1; s < s_max; ++s){
	  if(s == n) continue;
	  result += h<-1>(n, s, p)*h_light<0>(n, s, xi_n)/(s*(n-s));
	}
	result += (
	    n*h<0>(n, n, p)*h_light<0>(n, n, xi_n)
	    + h<-1>(n, n, p)*(
		+ h_light<1>(n, n, xi_n)
		- h_light<0>(n, n, xi_n)*(7./2.+n*psi<0>(n))
	    )
	)/(n*n);
	result *= ((n%2)?1:-1)*sqrt(pi)/(4*Gamma(n));
	return  pow(p.alpha_s/pi,2)*(
	    result
	  - delta_e1_Coulomb(n, p)*delta_f1_light(n, xi_n)
	  - delta_f1_Coulomb(n, p)*delta_e1_light(n, xi_n)
	);
      }

      /***************************************************
       *       NNNLO corrections to binding energy        *
       ***************************************************/

      /* correction to e3 from one insertion of coulomb potential */
      double delta_e3_Coulomb(int n, parameters const & p){
	using Wilson_coefficients::a;
	return pow(p.alpha_s/pi,3)/(64*n)*(
	    + 2*a<3>(p.nl)*n
	    + 36*zeta_2*a<1>(p.nl)*n*pow(4*beta(0, p.nl),2)
	    + 30*zeta_2*n*4*beta(0, p.nl)*16*beta(1, p.nl)
	    - 96*pow(4*beta(0, p.nl),3)*harmonic_sum<1, 1>(n)
	    - 96*n*pow(4*beta(0, p.nl),3)*harmonic_sum<2, 1>(n)
	    + 96*n*pow(4*beta(0, p.nl),3)*harmonic_sum<1, 1, 1>(n)
	    + 192*zeta_2*pow(c_A,3)*n*log_n(n, p)
	    + 12*a<2>(p.nl)*n*4*beta(0, p.nl)*log_n(n, p)
	    + 72*zeta_2*n*pow(4*beta(0, p.nl),3)*log_n(n, p)
	    + 8*a<1>(p.nl)*n*16*beta(1, p.nl)*log_n(n, p)
	    + 4*n*64*beta(2, p.nl)*log_n(n, p)
	    + 24*a<1>(p.nl)*n*pow(4*beta(0, p.nl)*log_n(n, p),2)
	    + 20*n*4*beta(0, p.nl)*16*beta(1, p.nl)*pow(log_n(n, p),2)
	    + 16*n*pow(4*beta(0, p.nl)*log_n(n, p),3)
	    + 4*4*beta(0, p.nl)*pow(harmonic_sum<1>(n),2)*(
		+ 6*a<1>(p.nl)*n*4*beta(0, p.nl)
		+ 12*pow(4*beta(0, p.nl),2)
		+ 5*n*16*beta(1, p.nl)
		+ 12*n*pow(4*beta(0, p.nl),2)*log_n(n, p)
	    )
	    + 4*4*beta(0, p.nl)*harmonic_sum<2>(n)*(
		+ 6*a<1>(p.nl)*n*4*beta(0, p.nl)
		+ 12*pow(4*beta(0, p.nl),2)
		+ 5*n*16*beta(1, p.nl)
		+ 12*n*pow(4*beta(0, p.nl),2)*log_n(n, p)
	    )
	    + 4*harmonic_sum<1>(n)*(
		+ 48*zeta_2*pow(c_A,3)*n
		+ 3*a<2>(p.nl)*n*4*beta(0, p.nl)
		- 12*a<1>(p.nl)*pow(4*beta(0, p.nl),2)
		+ 18*zeta_2*n*pow(4*beta(0, p.nl),3)
		+ 2*a<1>(p.nl)*n*16*beta(1, p.nl)
		- 10*4*beta(0, p.nl)*16*beta(1, p.nl)
		+ n*64*beta(2, p.nl)
		+ 2*4*beta(0, p.nl)*(
		    + 6*a<1>(p.nl)*n*4*beta(0, p.nl)
		    - 12*pow(4*beta(0, p.nl),2)
		    + 5*n*16*beta(1, p.nl)
		)*log_n(n, p)
		+ 12*n*pow(4*beta(0, p.nl),3)*pow(log_n(n, p),2)
	    )
	);
      }

      /* correction to e3 from two insertions of coulomb potential */
      double delta_e3_double_Coulomb(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	using Wilson_coefficients::a;
        const double VC2_log_coeff =
          2*a<1>(p.nl)*4*beta(0, p.nl) + 16*beta(1, p.nl)
          ;
	return pow(p.alpha_s/pi,3)/n*(
            + a1_eff(p)*n*(
                + a<2>(p.nl)/32
                + VC2_log_coeff*(
                    -1./16. + log_n(n, p)/16 + harmonic_sum<1>(n)/16
                )
            )
            + 4*beta(0, p.nl)*(
                VC2_log_coeff*harmonic_sum<1>(n)/4
                + 4*beta(0, p.nl)*(
                    (
                        - a1_eff(p)/4
                        + 4*beta(0, p.nl)*(
                            1./2. + log_n(n, p)/2 - harmonic_sum<1>(n)
                        )
                    )*harmonic_sum<1>(n) - 2*beta(0, p.nl)*harmonic_sum<2>(n)
                )
                + n*(
                    a<2>(p.nl)*(
                        - 1./16. + log_n(n, p)/16 + harmonic_sum<1>(n)/16
                    )
                    + VC2_log_coeff*(
                        - zeta_2/4.
                        + log_n(n, p)*(
                            - 1./4. + log_n(n, p)/8 + harmonic_sum<1>(n)/4
                        )
                        + (- 1./4. + harmonic_sum<1>(n)/8)*harmonic_sum<1>(n)
                        + harmonic_sum<2>(n)/4
                        + n*(zeta_3/4. - harmonic_sum<3>(n)/4)
                    )
                    + 4*beta(0, p.nl)*(
                        + a1_eff(p)*(
                            3./16.*zeta_2
                            + log_n(n, p)*(
                                -1./4. + log_n(n, p)/8 + harmonic_sum<1>(n)/4
                            )
                            + (-1./4. + harmonic_sum<1>(n)/8)*harmonic_sum<1>(n)
                            + harmonic_sum<2>(n)/8
                        )
                        + 4*beta(0, p.nl)*(
                            -3./8.*zeta_2 + zeta_3
                            + harmonic_sum<1>(n)*(
                                3./8.*zeta_2
                                + (
                                    - 3./4. + harmonic_sum<1>(n)/4
                                )*harmonic_sum<1>(n)
                                + harmonic_sum<2>(n)/4
                            )
                            - harmonic_sum<2>(n)/4
                            + log_n(n, p)*(
                                - 5./8.*zeta_2
                                + log_n(n, p)*(
                                    - 3./4.
                                    + log_n(n, p)/4
                                    + 3./4.*harmonic_sum<1>(n)
                                )
                                + (
                                    - 3./2.
                                    + 3./4.*harmonic_sum<1>(n)
                                )*harmonic_sum<1>(n)
                                + 5./4.*harmonic_sum<2>(n)
                                + n*(zeta_3 - harmonic_sum<3>(n))
                            )
                            + harmonic_sum<3>(n)
                            - harmonic_sum<2, 1>(n)
                            + n*(
                                - zeta_4/4.
                                - zeta_2*harmonic_sum<2>(n)
                                + harmonic_sum<1>(n)*(
                                    zeta_3 - harmonic_sum<3>(n)
                                )
                                - harmonic_sum<4>(n)
                                + 3*harmonic_sum<3, 1>(n)
                            )
                        )
                    )
                )
            )
        );
      }

      /* correction to e3 from three insertions of coulomb potential */
      double delta_e3_triple_Coulomb(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	return pow(p.alpha_s/pi,3)/(288*n)*4*beta(0, p.nl)*(
	    - 36*(-4+n)*pow(4*beta(0, p.nl)*harmonic_sum<1>(n),2)
	    + 36*pow(n*4*beta(0, p.nl)*harmonic_sum<2>(n),2)
	    - 12*4*beta(0, p.nl)*harmonic_sum<1>(n)*(
		+ 3*a1_eff(p)*n
		+ 6*4*beta(0, p.nl)*(1 - n*(1 - log_n(n, p)))
		- 12*n*4*beta(0, p.nl)*(harmonic_sum<2>(n) - zeta_2)
	    )
	    + 72*pow(4*beta(0, p.nl),2)*harmonic_sum<2>(n)*(
		+ 1 - n - zeta_2*pow(n,2)
		+ pow(n,3)*(harmonic_sum<3>(n) - zeta_3)
	    )
	    + n*(
		- 9*pow(a1_eff(p),2)
		+ 36*a1_eff(p)*4*beta(0, p.nl)*(1 - log_n(n, p))
		+ pow(4*beta(0, p.nl),2)*(
		    + 72*zeta_2
		    + 72*(harmonic_sum<2, 1>(n) -2*zeta_3)
		    + 36*log_n(n, p)*(2 - log_n(n, p))
		    + n*(
			+ (
			    72 + 216*zeta_2*n
			)*(harmonic_sum<3>(n) - zeta_3)
			+ 90*zeta_4
			- 180*harmonic_sum<4>(n)
			+ 144*harmonic_sum<3, 1>(n)
			+ n*(
			    + 216*harmonic_sum<5>(n)
			    - 144*harmonic_sum<3, 2>(n)
			    - 432*harmonic_sum<4, 1>(n)
			    + 432*zeta_5
			)
		    )
		)
	    )
	);
      }

      /* correction to e3 from double insertion of coulomb and delta potential */
      double delta_e3_Coulomb_delta(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	using Wilson_coefficients::v0;
	return pow(p.alpha_s,3)/pi/n*(
	    + 3./2.*a1_eff(p)*v0(p.nl)*pow(c_F,2)
	    - pow(c_F,2)*v0(p.nl)*4*beta(0, p.nl)*(
		+ 1
		- 3*log_n(n, p)
		+ 2*n*zeta_2
		- harmonic_sum<1>(n)
		- 2*n*harmonic_sum<2>(n)
	    )
	);
      }

      /* correction to e3 from double insertion of coulomb and 1/r^2 potential */
      double delta_e3_Coulomb_r2inv(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	using Wilson_coefficients::b;
	return pow(p.alpha_s,3)/pi/n*b<1>(p.nl)*c_F*(
	    - 1./2.*a1_eff(p)
	    - 4*beta(0, p.nl)*(
		- 1./2.
		+ log_n(n, p)
		- n*zeta_2
		+ harmonic_sum<1>(n)
		+ n*harmonic_sum<2>(n)
	    )
	);
      }

      /* correction to e3 from double insertion
	 of coulomb and momentum dependent potential */
      double delta_e3_Coulomb_p2(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	return pow(p.alpha_s,3)/(8*pow(n,2)*pi)*pow(c_F,2)*(
	    + 3*a1_eff(p)*(4*n - 1)
	    + 4*beta(0, p.nl)*(
		+ 2
		+ 6*log_n(n, p)*(4*n - 1)
		- 16*zeta_2*pow(n,2)
		+ (8*n - 6)*harmonic_sum<1>(n)
		+ 16*pow(n,2)*harmonic_sum<2>(n)
	    )
	);
      }

      /* correction to e3 from double insertion
	 of coulomb potential and kinetic correction */
      double delta_e3_Coulomb_kinetic(int n, parameters const & p){
	using Wilson_coefficients::a1_eff;
	return pow(p.alpha_s,3)/(16*pi*pow(n,2))*pow(c_F,2)*(
	    + a1_eff(p)*(8*n-3)
	    + 4*beta(0, p.nl)*(
		+ 2*(1 + 2*n)
		+ 2*log_n(n, p)*(8*n-3)
		- 8*zeta_2*pow(n,2)
		- 6*harmonic_sum<1>(n)
		+ 8*pow(n,2)*harmonic_sum<2>(n)
	    )
	);
      }

      /* correction to e3 from insertion of delta potential */
      double delta_e3_delta(int n, parameters const & p){
	return pow(p.alpha_s,3)/pi/pow(n,2)*(
	    T_f*pow(c_F,2)*n*(2./15. + 20./27.*p.nl)
	    +pow(c_F,3)*(
		+ 7./6.
		- 3./2.*n
		+ 8./3.*n*log_n(n, p)
		+ log(c_F*p.alpha_s/n)/3*n
		- 7./3.*n*harmonic_sum<1>(n)
	    )
	    + pow(c_F,2)*4*beta(0, p.nl)*(
		+ 1./3.*(n - 1)
		+ 2./3.*n*(harmonic_sum<1>(n) - log_n(n, p))
	    )
	    + c_A*pow(c_F,2)*(
		- 11./12.
		- 97./108.*n
		- 4./3.*n*log_n(n, p)
		+ log(c_F*p.alpha_s/n)/2*n
		+ 11./6.*n*harmonic_sum<1>(n)
	    )
	);
      }

      /* correction to e3 from insertion of 1/r^2 potential */
      double delta_e3_r2inv(int n, parameters const & p){
	using Wilson_coefficients::b;
	return pow(p.alpha_s,3)/pi/n*c_F*(
	    - b<2>(p.nl)
	    + (8./3.*c_A*(c_A+2*c_F) - 4*beta(0, p.nl)*b<1>(p.nl))*(
		1 + log_n(n, p) - harmonic_sum<1>(n)
	    )
	);
      }

      /* correction to e3 from insertion of momentum dependent potential */
      double delta_e3_p2(int n, parameters const & p){
	return pow(p.alpha_s,3)/pi/pow(n,2)*(
	    pow(c_F,2)*T_f*p.nl*(
		5./18. -10./9.*n
	    )
	    + c_A*pow(c_F,2)*(
		+ 2./3.*log_n(n, p)*(4*n - 1)
		- 31./72.
		+ 79./18.*n
		- 2./3.*harmonic_sum<1>(n)*(4*n + 1)
	    )
	    + pow(c_F,2)*4*beta(0, p.nl)*(
		+ 1./4.*log_n(n, p)*(4*n - 1)
		+ n
		- 1./4.*harmonic_sum<1>(n) * (4*n + 1)
	    )
	);
      }

      double delta_e3_Higgs(int n, parameters const & p){
	return c_F*p.alpha_s/(4*pi)*p.Yukawa_t2*pow(p.m/p.m_Higgs,2)/n;
      }

      /* ultra-soft correction to e3 */
      double log_bethe(int n){
	if((n < 1) || (n > 6)){
	  throw std::out_of_range(
	      "Bethe logarithm unknown for n = "
	      + std::to_string(n)
	      + " (not in range (1...6))"
	  );
	}
	static constexpr auto log_tab = std::array<double,6>{{
	    -81.5379,
	    -37.671,
	    -22.4818,
	    -14.5326,
	    -9.52642,
	    -6.0222,
	  }};
	return log_tab[n-1];
      }

      double delta_e3_ultrasoft(int n, parameters const & p){
	double log_as = log(c_F*p.alpha_s);
	double log_m = log(p.mu/p.m);
	return pow(p.alpha_s,3)/pi/pow(n,2)*(
	    - n*pow(c_F,3)*(
		+ 20./9.
		+ 8./3.*(log(2) + log_m)
		- 16./3.*log_as
		+ 2./3.*n*log_bethe(n)
	    )
	    + c_A*pow(c_F,2)*(
		+ 5./9.*(1 - 6*n)
		+ 2./3.*(log(2) + log_m)*(1 - 10*n)
		- 4./3.*log_as*(1 - 8*n)
		+ 8./3.*n*(harmonic_sum<1>(n) - log(n))
	    )
	    - n*pow(c_A,2)*c_F*(
		+ 25./9.
		- 4*log_as
		+ 8./3.*(log(2) + log_m)
		- 4./3.*(harmonic_sum<1>(n) - log(n))
	    )
	    - pow(n,2)*pow(c_A,3)*(
		+ 5./36.
		+ 1./6.*log(2)
		+ 1./2.*log_m
		- 2./3.*log_as
		+ 1./3.*(harmonic_sum<1>(n) + log(n))
	    )
	);
      }
    }
  }

}
