/**
 * @file
 * @brief Implementation of coefficients for double potential insertions
 *
 * All equation numbers refer to
 * Beneke, Maier, Piclum, Rauh, Nucl. Phys. B891 42-72, arXiv:1411.3132
 * \cite Beneke:2014pta
 *
 * @see h.hpp
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.3
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include "h.hpp"

#include <tuple>
#include <map>
#include <functional>

#include "Gamma.hpp"
#include "complex.hpp"
#include "psi.hpp"
#include "integrate.hpp"
#include "zeta.hpp"
#include "QCD/beta.hpp"
#include "Wilson_coefficients.hpp"
#include "log_n.hpp"

namespace QQbar_threshold{
  namespace detail{

    namespace{
      using namespace complex;
      typedef std::complex<double> complex;
      using Gamma::Gamma;
      using psi::psi;
      using QCD::beta;
      using harmonic_sums::harmonic_sum;
      using namespace zeta;

      //! I^{(1)}(u) as in eq. (C.14)
      complex I1(complex u){
	return Gamma(2-u/2)/(u*Gamma((5-u)/2));
      }

      //! \kappa as in eqs. (C.21)-(C.23)
      template<int j>
	complex kappa(int n, complex u, complex x);

      template<>
	complex kappa<0>(int, complex, complex){
	return 1;
      }

      template<>
	complex kappa<1>(int n, complex u, complex x){
	return u + n*psi<0>(x);
      }

      template<>
	complex kappa<2>(int n, complex u, complex x){
	return u*(u-1) + n*psi<0>(x)*(2*u + n*psi<0>(x)) - n*n*psi<1>(x);
      }

      //! Integrand from eq. (C.20)
      template<int j>
	double h_light_integrand(double y, int n, int k, double xi_n){
	static constexpr double del = 0.5;
	complex const u{del, log(y/(1-y))};
	return real(
	    pow(xi_n, -u)*I1(u)*(k+u)*pow(Gamma(k+u),2)*Gamma(-u)/Gamma(1+k-n+u)
	    *kappa<j>(n, u, 1+k-n+u)
	)/(y*(1-y));
      }

      //! Integral from eq. (C.20)
      /**
       * We cache the result to avoid calculating identical
       * integrals more than once
       */
      template<int j>
	double h_light_integral(int n, int k, double xi_n){
	auto index = std::make_tuple(n, k, xi_n);
	static std::map<decltype(index), double> lookup_table;

	auto it = lookup_table.find(index);
	if(it != lookup_table.end()) return it->second;

	using namespace std::placeholders;
	auto integrand = std::bind(h_light_integrand<j>, _1, n, k, xi_n);
	auto i = integrate::make_integrator(integrand);
	double result = i.integrate(0, 1, 1e-5, 1e-20)/(2*pi);
	lookup_table.emplace(index, result);
	return result;
      }
    }

    //! h^{(j)}_{m_c}(s) from eq. (C.20)
    template<int j>
    double h_light(int n, int s, double xi_n){
      double result = 0;
      for(int k=1; k <= s; ++k){
	result +=
	  ((k%2)?-1:1)*Gamma(s+1)/(k*pow(Gamma(k),2)*Gamma(s-k+1))
	  *h_light_integral<j>(n, k, xi_n)
	  ;
      }
      return result;
    }

    // force instantiation
    template double h_light<0>(int n, int s, double xi_n);
    template double h_light<1>(int n, int s, double xi_n);
    template double h_light<2>(int n, int s, double xi_n);

    //! h^{(-1)}_{massless}(s) from eqs. (C.26)-(C.28)
    template<>
    double h<-1>(int n, int s, parameters const & p){
      using Wilson_coefficients::a1_eff;
      if(s < n) return 2*4*beta(0, p.nl)*s/(n-s);
      if(s == n) {
	return -n*(
	    a1_eff(p) + 2*4*beta(0, p.nl)*(log_n(n, p) + harmonic_sum<1>(n))
	);
      }
      return 2*4*beta(0, p.nl)*n/(s-n);
    }

    //! h^{(0)}_{massless}(s) from eqs. (C.29)-(C.31)
    template<>
    double h<0>(int n, int s, parameters const & p){
      using Wilson_coefficients::a1_eff;
      if(s < n){
	return (
	- a1_eff(p)*s
	 + 2*4*beta(0, p.nl)*(
	     n*(psi<0>(n-s) - psi<0>(n))
	    + s*(harmonic_sum<1>(n-s) - log_n(n, p))
         )
	 - h<-1>(n, s, p)
      )/(n-s);
      }
      if(s == n) {
	return -2*4*beta(0, p.nl)*(1 + harmonic_sum<1>(n-1) + n*psi<1>(n));
      }
      return (
      + a1_eff(p)*s
	  - 2*4*beta(0, p.nl)*(
	      n*(psi<0>(s-n) - psi<0>(n))
	      + s*(harmonic_sum<1>(s-n) - log_n(n, p))
	      - 2
	  )
	  + h<-1>(n, s, p)
      )/(s-n);
    }

    //! h^{(1)}_{massless}(s) from eq. (C.32)
    template<>
    double h<1>(int n, int s, parameters const & p){
      (void) s;
      assert(n == s);
      return 4*beta(0, p.nl)*(
	  1./n + 2*harmonic_sum<2>(n-1) - n*(psi<2>(n) + 4*zeta_3)
      );
    }

    // force instantiation
    template double h<-1>(int n, int s, parameters const & p);
    template double h<0>(int n, int s, parameters const & p);
    template double h<1>(int n, int s, parameters const & p);

  }
}
