/**
 * @file
 * @brief Implementation of cross section functions
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.3
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include <cmath>
#include <complex>
#include <exception>
#include <vector>
#include <cassert>

#include "xsection.hpp"
#include "complex.hpp"
#include "mass_schemes.hpp"
#include "parameters.hpp"
#include "scheme_conversion.hpp"
#include "alpha_s.hpp"
#include "QCD/colour.hpp"
#include "constants.hpp"
#include "zeta.hpp"
#include "p_wave.hpp"
#include "Green_functions.hpp"
#include "pole_resummation.hpp"
#include "non-resonant.hpp"

namespace QQbar_threshold{
  using QCD::c_F;
  using namespace zeta;

  namespace detail{
    double QQbar_xsection(parameters const & p);
    double R_Q(parameters const & p);

    complex Pi_V(complex lam, parameters const & p);
    complex Pi_A(complex lam, parameters const & p);
    complex Pi_V_TM(complex lam, parameters const & p);
    complex Pi_A_TM(complex lam, parameters const & p);
    double mass_shift_to_energy(parameters const & p);

  }

  double ttbar_xsection(
      double sqrt_s, scales mu, top_properties M, order o,
      options opt
  ){
    return detail::QQbar_xsection(
        gen_top_parameters(sqrt_s, mu, M, o, opt)
    );
  }

  double ttbar_xsection(
      double sqrt_s, double mu, top_properties M, order o,
      options opt
  ){
    return detail::QQbar_xsection(
        gen_top_parameters(sqrt_s, mu, M, o, opt)
    );
  }

  double bbbar_xsection(
      double sqrt_s, double mu, double m, order o,
      options opt
  ){
    return detail::QQbar_xsection(
        gen_bottom_parameters(sqrt_s, mu, m, o, opt)
    );
  }

  double ttbar_threshold(
      double mu, double m, order o,
      options opt
  ){
    if(is_shifted_mass(opt.mass_scheme)){
      return 2*(m + detail::top_mass_shift(mu, m, o, std::move(opt)));
    }
    if(is_threshold_mass(opt.mass_scheme)){
      return 2*(m + detail::delta_m(0, gen_top_parameters(mu, m, o, std::move(opt))));
    }
    return 2*m;
  }

  double bbbar_threshold(
      double mu, double m, order o,
      options opt
  ){
    if(is_shifted_mass(opt.mass_scheme)){
      return 2*(m + detail::bottom_mass_shift(mu, m, o, std::move(opt)));
    }
    if(is_threshold_mass(opt.mass_scheme)){
      return 2*(m + detail::delta_m(0, gen_bottom_parameters(mu, m, o, std::move(opt))));
    }
    return 2*m;
  }

  double ttbar_R_ratio(
      double sqrt_s, scales mu, top_properties M, order o,
      options opt
  ){
    return detail::R_Q(
        gen_top_parameters(sqrt_s, mu, M, o, opt)
    );
  }

  double ttbar_R_ratio(
      double sqrt_s, double mu, top_properties M, order o,
      options opt
  ){
    return detail::R_Q(
        gen_top_parameters(sqrt_s, mu, M, o, opt)
    );
  }

  double bbbar_R_ratio(
      double sqrt_s, double mu, double m, order o,
      options opt
  ){
    return detail::R_Q(
        gen_bottom_parameters(sqrt_s, mu, m, o, opt)
    );
  }

  namespace detail{

    // Main functions for cross section

    double resonant_xsection(parameters const & p);
    double normalised_xsection(parameters const & p);

    double QQbar_xsection(parameters const & p){
      return invGeV2_to_pb*4*pi/3*p.alpha*p.alpha/p.s*R_Q(p);
    }

    double R_Q(parameters const & p){
      if(p.resonant_only || (p.order == 0)){
        return resonant_xsection(p);
      }
      return resonant_xsection(p) + nonresonant_xsection(p);
    }

    double S_wave_xsection_EW(complex lam, parameters const & p){
      using Wilson_coefficients::cv;
      using Wilson_coefficients::Cv_p;
      using Wilson_coefficients::Ca_p;
      using Wilson_coefficients::Cv_EW;
      using Wilson_coefficients::Ca_EW;
      if(p.order < 2) return 0;
      auto Green_fun = std::vector<complex>(p.order-1);

      switch(p.order){
      case 3:
	Green_fun[1] = Green_function_PR_mu_weak_in_real<1>(lam,p);
      default:
	Green_fun[0] = Green_function_PR_mu_weak_in_real<0>(lam,p);
      }
      complex result = Green_fun[0];
      if(p.order == 3){
	result +=
	  + cv<1>(p) * Green_fun[0]
	  + Green_fun[1];
      }
      result *= (p.production == production_channel::photon_only)?
        (Cv_p(p)*Cv_EW<2>(p)):
        (Cv_p(p)*Cv_EW<2>(p) + Ca_p(p)*Ca_EW<2>(p))
        ;
      result *= p.expand_s?(3*QCD::n_c/(p.m*p.m)):(12*QCD::n_c/p.s);
      return p.alpha*imag(result);
    }

    double S_wave_xsection_QCD(complex lam, parameters const & p){
      using Wilson_coefficients::Cv_p;
      using Wilson_coefficients::Ca_p;
      double const R_V = 12*pi*imag(Pi_V(lam,p));
      double const Cv = Cv_p(p);
      if(p.production == production_channel::photon_only) return Cv*Cv * R_V;
      double const Ca = Ca_p(p);
      return (Cv*Cv + Ca*Ca) * R_V;
    }

    double S_wave_xsection(complex lam, parameters const & p){
      if(!(p.beyond_QCD & SM::QED)){
	return S_wave_xsection_QCD(lam, p);
      }
      return S_wave_xsection_QCD(lam, p) + S_wave_xsection_EW(lam, p);
    }

    double P_wave_xsection(complex lam, parameters const & p){
      double const R_A = 12*pi*imag(Pi_A(lam,p));
      double const a_q = p.c_q[axial];
      static const double a_e = c_e<axial>();
      static const double v_e = c_e<vector>();
      return R_A * (a_e*a_e + v_e*v_e)*pow(a_q*p.s/(p.s-mZ*mZ), 2);
    }

    double resonant_xsection(parameters const & p){
      complex const lam = p.alpha_s*c_F/2. * sqrt(-p.m/p.E_shifted);
      if(p.production == production_channel::all){
	return S_wave_xsection(lam, p) + P_wave_xsection(lam, p);
      }
      return S_wave_xsection(lam, p);
    }

    //! correction factor from (possible) expansion of 1/s
    /*
     * In principle terms that come from the expansion of the factor 1/s
     * in the polarisation function should be multiplied by the
     * unshifted energy, i.e. 1/s = 1/(4*m^2)*(1 - E/m + ...), where
     * E is the energy *without* shift.
     * However, there is an additional shift from the overall factor m^2
     * multiplying the Green function. This shift is *not* accounted for in the
     * threshold mass corrections GreenFunctionTM and has to be treated here.
     * If we do not expand 1/s it is all that remains.
     */
    double get_s_exp_factor(parameters const & p){
      if(p.expand_s){
	return - real(p.E_shifted);
      }
      if(is_threshold_mass(p.mass_scheme)) return 2*delta_m(0, p);
      return 0;
    }

    //! Vector polarisation function
    complex Pi_V(complex lam, parameters const & p){
      if((p.order < 0) || (p.order > 3)){
	throw std::out_of_range(
	    "order "+ std::to_string(p.order)
	    + " out of range (0,...,3)"
	);
      }
      auto Green_fun = std::vector<complex>(p.order+1);
      auto cv = std::vector<double>(p.order+1);
      auto dv = std::vector<double>(std::max(p.order-1,0));
      switch(p.order){
      case 3:
      	Green_fun[3] = Green_function<3>(lam,p);
      	cv[3] = Wilson_coefficients::cv<3>(p);
      	dv[1] = Wilson_coefficients::dv<1>(p);
      case 2:
      	Green_fun[2] = Green_function<2>(lam,p);
      	cv[2] = Wilson_coefficients::cv<2>(p);
      	dv[0] = Wilson_coefficients::dv<0>(p);
      case 1:
      	Green_fun[1] = Green_function<1>(lam,p);
      	cv[1] = Wilson_coefficients::cv<1>(p);
      default:
      	Green_fun[0] = Green_function<0>(lam,p);
      }
      complex const & E_shifted = p.E_shifted;
      double const & m =p.m;
      complex const s_exp_factor = get_s_exp_factor(p);

      complex result = Green_fun[0];
      switch(p.order){
      case 3:
      	result += (
         Green_fun[0]*(
            2*cv[1]*cv[2] +
            2*cv[3] -
            E_shifted/m*(
               1./3.*dv[1] +
               cv[1]*1./3.*dv[0]
            )
	    + s_exp_factor/m*2*cv[1]
         ) +
         Green_fun[1]*(
            pow(cv[1],2) +
            2*cv[2] -
            E_shifted/m*1./3.*dv[0]
	    + s_exp_factor/m
         ) +
         Green_fun[2]*2*cv[1] +
         Green_fun[3]
	 + complex{0, pi*c_F*p.m*p.width/p.alpha_s}*Green_function_weak<1>(p)
      	);
      case 2:
      result += (
         Green_fun[0]*(
            pow(cv[1],2) + 2*cv[2] -
            E_shifted/m*1./3.*dv[0]
	    + s_exp_factor/m
         )+
         Green_fun[1]*2*cv[1] +
         Green_fun[2]
	 + complex{0, pi*c_F*p.m*p.width/p.alpha_s}*Green_function_weak<0>(p)
      );
      case 1:
      	result += (
	    + Green_fun[0]*2*cv[1]
	    + Green_fun[1]
      	);
      default:;
      }
      if(p.expand_s){
      	result *= QCD::n_c/(2*pow(m,2));
      }
      else {
      	result *= 2*QCD::n_c/p.s;
      }
      for(int n = 1; n <= p.resum_poles; ++n){
	result += QCD::n_c/(2*pow(m,2))*resum_pole_Pi(n, p);
      }
      if(is_threshold_mass(p.mass_scheme)){
	result += Pi_V_TM(lam, p);
      }
      return result;
    }

    //! Axial-vector polarisation function
    complex Pi_A(complex lam, parameters const & p){
      using QCD::c_F;
      using Wilson_coefficients::a1_eff;
      using Wilson_coefficients::ca;
      if(p.order < 2) return 0;
      if(p.order > 3){
	throw std::out_of_range(
	    "order "+ std::to_string(p.order)
	    + " out of range (0,...,3)"
	);
      }
      const double log_m = log(p.m/p.mu_w);
      complex result =
	+ Green_function_P<0>(lam,p)
	//- pow(c_F*p.m*p.alpha_s/lam,3)*p.m/(864*pi)*(pow(lam,3)/4 - lam/2)
	- c_F*p.alpha_s*pow(p.m,3)/(384*pi)*(
	    8*p.E_shifted + pow(c_F*p.alpha_s,2)*p.m
	)
	;
      if(p.order > 2) {
	result += (
	    + Green_function_P<1>(lam,p)
	    + 2*ca<1>(p)*Green_function_P<0>(lam,p)
	    + c_F*pow(p.m,3)*p.alpha_s*(
		- p.E_shifted*ca<1>(p)/(24.*pi)
		+ p.alpha_s*(
		    - pow(c_F,2)/(192.*pi)*ca<1>(p)*p.m*p.alpha_s
		    + pow(pi,-2)*(
			p.E_shifted*p.contributions.v_Coulomb[0]*(
			    - a1_eff(p)/288.
			    + 4./864.*beta(0, p.nl)
			)
			+ c_F*(
			    + p.contributions.ca[0]*p.E_shifted*log_m/4.
			    + c_F*p.m*pow(p.alpha_s,2)*(
                                + p.contributions.v_Coulomb[0]*(
                                    - a1_eff(p)/640. + 4./1920.*beta(0, p.nl)
                                )
				+ 9./278.*c_F*p.contributions.ca[0]*log_m
			    )
			)
		    )
		)
	    )
	);
      }
      if(p.expand_s){
	result *= QCD::n_c/3.*pow(p.m, -4);
      }
      else {
	result *= 4./3.*QCD::n_c*1/(p.s*pow(p.m,2));
      }
      if(p.resum_poles){
        result += pole_resummation_P(lam, p);
      }
      if(is_threshold_mass(p.mass_scheme)){
	result += Pi_A_TM(lam, p);
      }

      return result;
    }

    //! Threshold mass corrections to the vector polarisation function
    /**
     * I.e. Corrections to the polarisation function from expanding the mass
     * around the leading-order relation to the pole mass. Only relevant for
     * threshold masses in the "insertion" variant of the scheme.
     */
    complex Pi_V_TM(complex lam, parameters const & p){
      using Wilson_coefficients::cv;
      using Wilson_coefficients::dv;
      assert((p.order >= 0) && (p.order < 4));
      if(p.order == 0) return 0;
      double const & m =p.m;
      complex const s_exp_factor = get_s_exp_factor(p);

      complex result = Green_function_TM<0,1>(lam, p);
      switch(p.order){
      case 3:
	result +=
	  + Green_function_TM<0,3>(lam, p)
	  + Green_function_TM<1,2>(lam, p)
	  + Green_function_TM<2,1>(lam, p)
	  + 2*cv<1>(p)*(
	      + Green_function_TM<0,2>(lam, p)
	      + Green_function_TM<1,1>(lam, p)
	  )
	  + Green_function_TM<0,1>(lam, p)*(
	      + 2*cv<2>(p)
	      + pow(cv<1>(p),2)
	      - p.E_shifted/m*1./3.*dv<0>(p)
	      + s_exp_factor/m
	  )
	  - mass_shift_to_energy(p)/m * Green_function<0>(lam,p) * (
	      1 + 1./3.*dv<0>(p)
	  )
          ;
      case 2:
	result +=
	  + Green_function_TM<0,2>(lam, p)
	  + Green_function_TM<1,1>(lam, p)
	  + 2*cv<1>(p)*Green_function_TM<0,1>(lam, p)
	  ;
      default:;
      }
      if(p.expand_s){
	result *= QCD::n_c/(2*pow(m,2));
      }
      else {
	result *= 2*QCD::n_c/p.s;
      }
      return result;
    }

    //! Threshold mass corrections to the axial-vector polarisation function
    complex Pi_A_TM(complex lam, parameters const & p){
      assert((p.order >= 0) && (p.order < 4));
      if(p.order < 3) return 0;
      complex result =
	Green_function_P_TM<0,1>(lam, p)
	+ c_F/(216*pi)*p.alpha_s*pow(p.m,3)*delta_m(1, p)
	;
      if(p.expand_s){
	result *= QCD::n_c/3.*1/pow(p.m,4);
      }
      else {
	result *= 4./3.*QCD::n_c*1/(p.s*pow(p.m,2));
      }
      return result;
    }

    double mass_shift_to_energy(parameters const & p){
      if((!is_threshold_mass(p.mass_scheme))
	 || (p.order < 3)
      ) return 0;
      return -2*delta_m(1, p);
    }
  }

}
